#==============================================================================
#
#  $Id$
#
"""
   Python template processor.
"""
#
#  Copyright (C) 2005 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#==============================================================================

import re, sys

class ParseError(Exception):
   pass

class Processor:
   """
      A template processor instance.  This poorly designed class contains
      functionality to read and parse a template stream and execute the
      resulting template.
   """

   __defLineRx = re.compile(r'@')
   __defVarRx = re.compile(r'\$[a-z]\w*', re.I)
   __defStartExprRx = re.compile(r'\$\{')

   # the "default inline regex" covers both variables and nested expressions,
   # making it easier to search.  group 'var' is expected to be the name of a
   # variable if the match is a variable, group 'expr' is expected to be
   # non-empty if the variable is a nested expression.
   __defVarPat = r'\$(?P<var>[a-z]\w*)'
   __defStartExprPat = r'\$\{'
   __inlTmplt = r"(%s)|(?P<expr>%s)|(?P<escape>'''|\\)"
   __defInlRx = \
      re.compile(__inlTmplt % (__defVarPat, __defStartExprPat))
   __defEndExprRx = re.compile(r'\}\$')
   __defStartBlockRx = re.compile(r'@\s*\{')
   __defEndBlockRx = re.compile(r'@\s*\}')
   __defProcInsRx = re.compile(r'@!')

   # output states
   CODE = 0
   LITERAL = 1
   INLINE = 3

   # input states
   NORMAL = 0
   BLOCK = 1
   EXPR = 2
   PROCINS = 3
   
   # --- methods for use in processing instructions ---
   def restoreDefaultSymbols(self):
      self.lineRx = self.__defLineRx
      self.inlRx = self.__defInlRx
      self.endExprRx = self.__defEndExprRx
      self.startBlockRx = self.__defStartBlockRx
      self.endBlockRx = self.__defEndBlockRx
      self.procInsRx = self.__defProcInsRx

   def setSymbols(self, beginBlock = None, endBlock = None, statement = None,
		  variable = None,
		  beginExpr = None, 
		  endExpr = None, 
		  procIns = None
		  ):
      if beginBlock:
	 self.startBlockRx = re.compile(beginBlock)

      if endBlock:
	 self.endBlockRx = re.compile(endBlock)

      if statement:
	 self.lineRx = re.compile(statement)
      
      # we have to combine all of the inline patterns, check for them
      # independently after setting up to use their defaults
      brx = self.__defStartExprPat
      vrx = self.__defVarPat
      if beginExpr:
	 brx = beginExpr
      if variable:
	 vrx = variable
      if beginExpr or variable:
	 self.inlRx = re.compile(self.__inlTmplt % (vrx, brx))

      if endExpr:
	 self.endExprRx = re.compile(endExpr)

      if procIns:
	 self.procInsRx = re.compile(procIns)


   def __init__(self):
      self.restoreDefaultSymbols()

      self.__lineNo = 1
      self.__indent = 0
      self.__inputState = self.NORMAL
      self.state = self.CODE
      self.code = []
      self.__procInsContext = {
	 'setSymbols': self.setSymbols,
	 'restoreDefaultSymbols': self.restoreDefaultSymbols,
      }
      self.__lastProcIns = ''

   def __indentLine(self, line):
      return '  ' * self.__indent + line

   def __processProcIns(self, line, end):
      if self.state == self.LITERAL:
	 self.code[-1] = self.code[-1][:-1] + "\\n''')\n"
	 self.state = self.CODE

      # emit the line as a comment to preserve line numbering
      self.code.append("#" + line)

      # get the meat of the line and add it to the last processing instruction
      code = line[end:].strip()
      self.__lastProcIns += code + '\n'
 
      # if there's no continuation character...
      if code[-1] != '\\':
	 # execute it in the processing instruction context
	 try:
	    exec self.__lastProcIns in self.__procInsContext
	 except Exception, ex:
	    raise ParseError('%s:%d: error processing line: %s' %
			      (self.__srcName, self.__lineNo, ex)
			     )

	 # reset the processing instruction buffer
	 self.__lastProcIns = ''
	 self.__inputState = self.NORMAL
      else:
	 self.__inputState = self.PROCINS

   def __processCodeLine(self, line, end):
      # if we are not in CODE state, put us there.  We have to close off on
      # the _previous line_ in order to preserve the line-to-line mapping
      if self.state == self.LITERAL:
	 self.code[-1] = self.code[-1][:-1] + "\\n''')\n"
	 self.state = self.CODE

      # extract the meat of the line
      line = line[end:].strip()
      if line == '}':
	 # if the line is a single closing bracket, back out one indent level
	 indentDx = -1
	 line = '# }'
      elif line[-1] == '{':
	 # if the line ends with a bracket, indent one level and convert the
	 # bracket to a colon
	 indentDx = 1
	 line = line[:-1] + ':'
      else:
	 indentDx = 0

      # emit the code line
      self.code.append(self.__indentLine(line) + '\n')

      # apply the indent
      self.__indent += indentDx
      if self.__indent < 0:
         raise ParseError('%s:%d: Unmatched closing bracket' % 
                           (self.__srcName, self.__lineNo)
                          )

   def __emitInline(self, text):
      if self.state == self.LITERAL:
	 self.code.append("''' + ")
      elif self.state == self.CODE:
	 self.code.append(self.__indentLine('out.write('))
      else:
	 self.code.append(' + ')
      self.code.append(text)
      self.state = self.INLINE

   def __emitLiteral(self, text):
      if self.state == self.CODE:
	 self.code.append(self.__indentLine("out.write('''"))
      elif self.state == self.INLINE:
	 self.code.append(" + '''")

      self.code.append(text)
      self.state = self.LITERAL
 
   def __processExprLine(self, line):
      m = self.endExprRx.search(line)
      if m:
	 # we found a terminator - emit the line up to the terminator and hand
	 # the rest of it off to the normal line processor
	 self.code.append(line[:m.start()] + ')')
	 self.__inputState = self.NORMAL
	 self.__processNormalLine(line[m.end():])
      else:
	 self.code.append(line)

   def __processNormalLine(self, line):

      # check for an inline form
      m = self.inlRx.search(line)
      while m:

	 # write out the text before the form
	 if m.start() > 0:
	    self.__emitLiteral(line[:m.start()])

	 # is it a simple variable?
	 var = m.group('var')
	 if var:
	    self.__emitInline('str(%s)' % var)
	 elif m.group('expr'):
	    # switch to expression state, let the expression parser take care
	    # of the rest of the line
	    self.__inputState = self.EXPR
	    self.__emitInline('str(')
	    self.__processExprLine(line[m.end():])
	    return
	 else:
	    # special chars - need to escape
	    self.__emitLiteral("\\" + m.group())

	 # truncate what we've processed so far and check for another match
	 line = line[m.end():]
	 m = self.inlRx.search(line)

      # write the remainder of the line
      self.__emitLiteral(line)

   def read(self, src, name):
      self.__srcName = name
      for line in src:
	 # figure out how to proceed based on the "input state"

	 if self.__inputState == self.BLOCK:
	    # we're in block mode, check for an end block line
	    if self.endBlockRx.match(line):
	       self.code.append('#' + line)
	       self.state = self.CODE
	       self.__inputState = self.NORMAL
	    else:
	       self.code.append(self.__indentLine(line))
	 elif self.__inputState == self.EXPR:
	    self.__processExprLine(line)
	 elif self.__inputState == self.PROCINS:
	    # make sure we got a processing instruction prefix
	    m = self.procInsRx.match(line)
	    if m:
	       self.__processProcIns(line, m.end())
	    else:
	       raise ParseError('%s:%d: processing instruction continuation '
			         'expected' % (name, self.__lineNo)
			        )
	 else:
	    # normal state - check for special line indicators
	    if self.startBlockRx.match(line):

	       # start of a block
	       if self.state == self.LITERAL:
		  self.code.append("''') # " + line)
		  self.state = self.CODE
	       else:
		  self.code.append('#' + line)
	       self.__inputState = self.BLOCK
	       self.__lineNo += 1
	       continue

	    m = self.procInsRx.match(line)
	    if m:
	       # processing instruction
	       self.__processProcIns(line, m.end())
	       self.__lineNo += 1
	       continue

	    m = self.lineRx.match(line)
	    if m:
	       # code line
	       self.__processCodeLine(line, m.end())
	       self.__lineNo += 1
	       continue

	    # normal line
	    self.__processNormalLine(line)

	 self.__lineNo += 1

      # if we ended on a literal, close it off
      if self.state == self.LITERAL:
	 self.code.append("''')\n")

      # make sure that the parens are balanced (current indent should be zero)
      if self.__indent:
         raise ParseError('Missing bracket at end of file')


   def runCode(self, filename, transFile = None, context = None):
      """
	 Runs the code parsed by @read().

	 parms:
	    transFile::
	       [file.write or None] if present, this is an intermediate file
	       to write the processed code to.
	    context::
	       [dict<string, any> or None]
      """

      # concatenate the code list to the complete translation
      translation = ''.join(self.code)

      # write if requested
      if transFile:
	 transFile.write(translation)

      # if there is no context, construct one
      if not context:
	 context = {}

      # if there is no "out" variable in the context, use sys.stdout
      if not context.has_key('out'):
	 context['out'] = sys.stdout

      # compile and execute
      code = compile(''.join(self.code), filename, 'exec')
      exec code in context

def processFile(fileName, transFile = None, context = None):
   """
      Facade function to completely process an input file, optionally writing
      a translation file in the process.

      parms:
	 fileName::
	    [string] name of the input file.
   """
   src = open(fileName)
   mac = Processor()
   mac.read(src, fileName)
   mac.runCode(fileName, transFile, context)

