#==============================================================================
#
#  $Id: phatobj.py 467 1999-05-23 17:55:52Z mike $
#
"""
   Phat objects - objects that manage change and persistance.
"""
#
#  Copyright (C) 1999 Michael A. Muller
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#  $Log: phatobj.py,v $
#  Revision 1.1  1999/05/23 17:55:52  mike
#  First cut at phat objects
#
#  Revision 1.1  1999/02/19 00:35:06  mike
#  Initial revision
#
#
#==============================================================================

import thread
from pickle import Pickler, Unpickler

class SimpleIdSource:
   
   """
      Rules for an identification source:
      
      -  Id's must each have certain uniform syntactic characteristics,
         allowing different id namespaces to coexist.
         
      
      Thought about making identifiers predictable so that they could be
      used as part of a guaranteed delivery system, but I realize that this
      only complicates things:
      -  All transaction algorithms would have to be available
      -  It implies that guarenteed delivery is a product of the higher level
         systems
      Decided that guarenteed delivery should be an isolated feature.
      
      Current system uses sequence numbers.  At this time, we simply persist
      sequence numbers.  If the persistant number gets wiped out, we need
      to scan the database (and all mirror databases) for the last id from
      this source.
   """
   
   def createId(self, obj):
      """
         Creates a unique id for the object /obj/.
      """
      nextId = long(open('/home/mike/seqnum.dat').read())
      open('seqnum.dat', 'w').write(`nextId`)
      return 'POID-1.0-mmuller@enduden.com-' + `nextId` 

class Lockable:

   """
      Lockable objects maintain an internal semaphore which protects
      them when their state is volatile.
   """

   def __init__(self):
      self.__lock = thread.allocate_lock()
      
   def lockObject(self):
      """
         Locks an object, preventing any modifications from occurring.
      """
      self.__lock.acquire()

   def unlockObject(self):
      """
         Unlocks an object, allowing subsequent modifications.
      """
      self.__lock.release()

class Identifiable:
   def __init__(self):
      self.__id = _idsource.createId(self)

   def getId(self):
      return self.__id

class Persistant(Lockable, Identifiable):
   
   """
      A Persistant object is any object that implements the @storeOn()
      and @readFrom() methods.
      method.
   """

   def __init__(self):
      Lockable.__init__(self)
      Identifiable.__init__(self)
   
   def dict(self):
      """
         Returns the dictionary representation of an instance, stripped of
         all of it's non-persistant attributes.
         
         Derived classes wishing to override the __getstate__() method
         should override this instead, being sure to call this method from
         the overriding method.
      """
      dict = self.__dict__.copy()
      del dict['__getstate__']
      if dict.has_key('_Lockable__lock'):
         del dict['_Lockable__lock'] 
      return dict

   def reset(self, state):
      """
         Sets the instances contents from the given /state/, which should
         be a dictionary.
         
         Derived classes wishing to override the __setstate__() method
         should override this instead, being sure to call this method from
         the overriding method.
      """
      self.__dict__.update(state)
   
   def __setstate__(self, state):
      self.__dict__.update(state)
      if not hasattr(self, '_Lockable__lock'):
         self._Lockable__lock = thread.allocate_lock()
   
   def flatten(self, out):
      """
         Flattens /self/ onto the file object /out/.
         
         When a phat object is pickled, it is pickled onto its ID only.
         When it is flattened, it's entire contents are pickled. 
      """
      self.lockObject()
      try:
         # phake out the pickler by temporarily turning the getstate
         # method into the _dict function, which returns the instances
         # dictionary.  This bypasses the class level __getstate__
         # method.
         self.__getstate__ = self.dict
         Pickler(out).dump(self)
         del self.__getstate__
      finally:
         self.unlockObject()

   def resurrect(self, src):
      """
         Resurrects the receiver from the source stream /src/.  This method
         is necessary because a Persistant object that is unpickled only
         contains its id.
         
         Before this method is called, /self/ should contain only the 
         attribute /phatId/.  After this method is called, this attribute
         will have been deleted, and all of the original attributes of the
         object will have replaced it.
         
         If /self/ does not have a /phatId/ attribute, this method will 
         silently return.
      """
      if not hasattr(self, 'phatId'):
         return

      self.lockObject()
      
      try:
         newInst = Unpickler(src).load()
         assert isinstance(newInst, self.__class__) and \
            self.phatId == newInst._Identifiable__id
         lock = self._Lockable__lock
         self.__dict__.update(newInst.__dict__)
         self._Lockable__lock = lock
         del self.phatId
      finally:
         self.unlockObject()

   def __getstate__(self):
      """
         Returns a string representation refering to only the object id.
      """
      return { 'phatId': self._Identifiable__id }

class PhatObject:
   """
      Phat objects have:
      
      -  A Unique identifier assigned at construction.
      -  Transaction processing capability.  
      -  A persistance mechanism.  Embedded PhatObjects are represented
         by their id.
      -  An associated set of subscribers.  Subscribers can be associated at
         the class and instance level (possibly at the application level)
      
      I probably want to break these features up into seperate base classes.
      
      Subscription
      ------------
      
      Window updates, data mirroring, and persistance should all be 
      accomplished through subscription.  A subscriber is basically just
      a suite of functions to be called whenever the object is modified or
      deleted.
      
      Persistance
      -----------
      
      My current thinking is that persistance should be accomplished via
      subscription, same as everything else.
      
      There should be some concept of bundling: putting data that will
      be accessed together in the same place in the database.

      Transactions
      ------------
      
      All creation, deletion and modification of objects should be
      accomplished through transactions.  Transactions are portable,
      authenticatable and (at this time) apply to a single object.

      I might need to be able to create subtransactions - transactions that
      are part of a greater parent transaction and don't get committed or
      destroyed until the parent transaction does.

      Need something to make sure that transactions don't loop back around
      to their source objects.  One (bad) way to do this is to track all
      transactions that are applied to an object.  A better way is for
      the external dispatcher to filter out transactions that originated from
      the prospective destination.

      Persistant Subscription
      -----------------------
      
      Certain classes of subscription are persistant, others are not.  
      A subscription list should be associated with each object which knows
      how to persist itself using only

      Demand Loading
      --------------
      
      Not sure how demand loading of objects fits in at this point.  Maybe
      that's an mix in class.

      Security
      --------
      
      Should be able to attach security constraints to specific objects,
      You should be able give certain groups certain privileges, and there
      needs to be an authentication agent available to authenticate
      transactions as they enter from an insecure channel.

      To handle this, a transaction must have a signature at some point.

      The Format Method
      -----------------
      
      This method formats the record as a string based on a standard format
      descriptor, e.g. "%(id)s %(lastName)s"  It will call the 
      _constructedFields method to create fields that need to be constructed
      if a KeyError is thrown while formatting the record.
      
      Code Sharing
      ------------
      
      My thoughts on this are very immature, there should be some facility
      for moving modules around.  I guess that a Module must itself be a 
      PhatObject, and that the data mirroring mechanism has to know about
      changes to the code.
      
      Maybe there needs to be a PhatCode object whose class subscribers
      may include the revision control system???
      
      Phuck that.  The phat object should be the underlying basis for the
      revision control system.  Transactions could be code patches.

            
   """

   def __init__(self):
      """
         Constructs a PhatObject.  Construction of the object automatically
         creates a transaction.
      """

   def createTransaction(self):
      """
         Creates a transaction so that you can begin modifying the object.
      """
   
   def commitTransaction(self):
      """
         Sends all changes to all interested subscribers.
      """
   
   def cancelTransaction(self):
      """
         Cancels all changes made since the last createTransaction().
      """

   def dataSize(self):
      """
         Returns the size of the object's flattened representation.
      """
   
   def flatten(self, out):
      """
         Flattens /self/ onto the file object /out/.
         
         When a phat object is pickled, it is pickled onto its ID only.
         When it is flattened, it's entire contents are pickled. 
      """

   def setPermissions(self, perm):
      """
         Sets the PermissionList for the object.
      """
   
   def getPermissions(self):
      """
         Returns the object's PermissionList.
      """
   
   def subscribe(self, subscriber):
      """
         Subscribes to all events published by the object.  /subscriber/
         should be a function which accepts a single Transaction instance.
      """

   def send(self, txn):
      """
         Sends the transaction to all subscribers.
      """

class Instance(PhatObject):
   """
   """

class List(PhatObject):
   """
   """

class Dict(PhatObject):
   """
   """

class Transaction:
   
   """
      Base class for all classes that modify a phat object.
      
      Derived classes perform a set of actions on an individual object.
      
      Actually, probably want transaction to just be an atomic bundle of 
      Actions.
      
      Might need to be able to perform a set of actions on a group of
      objects, but I don't think so.
   """      
   
   def __init__(self, obj):
      """
         Constructs a new transaction on the given object.
      """
      self.obj = obj

   def commit(self):
      """
         Commits the transaction, applying it to the object and performing
         all of the notifications.  
         
         Must be implemented by derived class
      """
   
   def __setattr__(self, attr, val):
      setattr(self.obj, attr, val)

_idsource = SimpleIdSource()   