#==============================================================================
#
#  $Id: FileBrowser.py 849 2004-10-15 17:33:09Z mike $
#
"""
   My kick-ass file browser.
"""
#
#  Copyright (C) 2004 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#==============================================================================

import os, stat
from fnmatch import fnmatch
from weakref import proxy
from Tkinter import *
from Controls import ScrollingList

FMT_BRIEF = 0
FMT_LONG = 1

class File:
   """
      Internal representation of a file.
      
      Public-vars:
         name::
            [string] unqualified file name
         dir::
            [string] containing directory name
         mode::
            [int] file mode
   """
   
   def __init__(self, name, dir):
      self.name = name
      self.dir = dir
      self.mode = os.stat(os.path.join(dir, name))[stat.ST_MODE]

   def fmt(self, format):
      """
         Returns the file info formatted to one of a number of specifications:
         
         `FMT_BRIEF`::
            just the unqualified name with a single character at the end
            to indicate file type.
         `FMT_LONG`::
            full "ls -l" style listing
      """
      if format:
         return '%s %s' % (self.modeAsStr(), self.name)
      else:
         if stat.S_ISDIR(self.mode):
            return self.name + '/'
         elif stat.S_ISFIFO(self.mode):
            return self.name + '|'
         else:
            return self.name

   def isDir(self):
      """
         Returns true if the file is a directory - or, more generall, if
         it can contain other file objects.
      """
      return stat.S_ISDIR(self.mode)

   REG = 1
   DIR = 2
   FIFO = 4
   CHR = 8
   BLK = 16
   LNK = 32
   SOCK = 64
   UNKNOWN = 128
   ALL = 127
   
   def getType(self):
      """
         Returns the file object type, which is an integer suitable
         for use in a bit set.
      """
      if stat.S_ISREG(self.mode):
         return self.REG
      elif stat.S_ISDIR(self.mode):
         return self.DIR
      elif stat.S_ISFIFO(self.mode):
         return self.FIFO
      elif stat.S_ISCHR(self.mode):
         return self.CHR
      elif stat.S_ISBLK(self.mode):
         return self.BLK
      elif stat.S_ISLNK(self.mode):
         return self.LNK
      elif stat.S_ISSOCK(self.mode):
         return self.SOCK
      else:
         return self.UNKNOWN

   def modeAsStr(self):
      # determine the first character, which indicates the type of thing
      if stat.S_ISREG(self.mode):
         mode = '-'
      elif stat.S_ISDIR(self.mode):
         mode = 'd'
      elif stat.S_ISFIFO(self.mode):
         mode = 'p'
      elif stat.S_ISCHR(self.mode):
         mode = 'c'
      elif stat.S_ISBLK(self.mode):
         mode = 'b'
      elif stat.S_ISLNK(self.mode):
         mode = 'l'
      elif stat.S_ISSOCK(self.mode):
         mode = 's'
      else:
         mode = '?'

      # iterate over the three sets of permission bits.  Each set has
      # its own "extension" modifiers - set uid, set gid, and sticky - 
      # which change the appearance of the execute flag if they are set
      for base, ext, mark in ((0100, 04000, 'Ss'), 
                             (010, 02000, 'Ss'), 
                             (1, 01000, 'Tt')):
         
         # if the modifier is set, the execFlag is one of the modifier flags.
         # if not, it is just the normal values
         if self.mode & ext:
            execFlag = mark
         else:
            execFlag = '-x'

         for act, flag in ((4, '-r'), (2, '-w'), (1, execFlag)):
            if self.mode & (act * base):
               mode += flag[1]
            else:
               mode += flag[0]
      
      return mode

class Filter:
   """
      Abstract base class for filter objects.
   """
   
   def __call__(self, file):
      """
         Must be implemented by derived classes, returns true if /file/ is
         selectable.
         
         parms:
            file::
               [@File] the object to query
      """
      raise NotImplementedError()

   def makeControl(self, parent):
      """
         Returns a widget representing the filter - should be the size of a
         button or entryfield.  Base class version returns a label containing
         the text "Filter".
      """
      return Label(parent, text = 'Filter')

class GlobFilter:
   """
      Simple name globbing filter.
   """
   
   def __init__(self, globs, types):
      """
         parms:
            globs::
               [list<string>] List of file-globbing patterns that will be
               accepted by this filter.
            types::
               [int] the set of file types (as returned by @File.getType())
               that are acceptable.
      """
      self.globs = globs
      self.types = types

   def __call__(self, file):
   
      # first check against the globs, if we don't match any we're done
      for expr in self.globs:
         if fnmatch(file.name, expr):
            break
      else:
         return False
      
      # verify that it is of an acceptable type
      return file.getType() & self.types

class FileBrowser(Frame):

   __normalChars = ''.join([chr(ch) for ch in range(32, 128)])
   
   def __init__(self, parent = None, width = 30, filename = None,
                filter = False,
                canCreate = False
                ):
      """
         parms:
            parent::
               [Frame] container window
            width::
               [int] entry field width
            filename::
               [string or None] the initial file name for the entryfield.  
               If *None*, the current directory is used.
            filter::
               [@Filter or None] If specified, this is an object that
               filters out what is displayed and selectable.
            canCreate::
               [boolean] if true, the user will be allowed to type in file 
               names that do not exist.  If false, the user will be
               given a bell if they try to type something that doesn't exist.
      """
      
      Frame.__init__(self, parent, class_ = 'SpugFileBrowser')

      selfProxy = proxy(self)
      
      # define the child controls
      
      row = 0
      self.__list = ScrollingList(selfProxy)
      self.__list.grid(row = row, column = 0, sticky = NSEW)
      self.__list.bind('<<ListboxSelect>>', selfProxy.__selected)
      row += 1
      
      self.__entry = Entry(selfProxy, width = width)
      self.__entry.grid(row = row, column = 0, sticky = NSEW)
      self.__entry.bind('<KeyPress>', selfProxy.__key)

      # set the internal parameters
      self.__filter = filter
      self.__canCreate = canCreate
            
      # use the current working directory if no filename is specified
      if filename == None:
         filename = os.getcwd()

      self.columnconfigure(0, weight = 1)
      self.rowconfigure(0, weight = 1)
      
      # switch that determines whether the entry is modified when the
      # selection changes
      self.__modifyEntry = True
      
      # default list format (see File.fmt() for details
      self.__format = FMT_BRIEF

      self.set(filename)      

   def changeFormat(self, format):
      """
         Changes the list format.  See @File.fmt() for details.
      """
      self.__format = format
      self.__show()
      if format == 0:
         self.__list.list['font'] = 'helvetica 8'
      else:
         self.__list.list['font'] = 'courier 8'

   def set(self, filename):
      self.__base, self.__selection = os.path.split(filename)
      
      # build the list of contents - this is the contents of the directory
      # with everything filtered out
      self.__contents = []
      for file in os.listdir(self.__base):
         file = File(file, self.__base)
         if file.isDir() or not self.__filter or self.__filter(file):
            self.__contents.append(file)
         
      # sort by name
      self.__contents.sort(lambda a, b: cmp(a.name, b.name))
      
      # reflect the new information in the user interface
      self.__show()

   def isValid(self):
      """
         Returns true if the current selection is valid.
      """
      if self.__canCreate:
         return self.__selection and True or False
      else:
         index = self.__findIndex(self.__selection)
         return index is not None and \
            self.__contents[index].name == self.__selection
   
   def get(self):
      """
         Returns the current selection (the unqualified name), a string.
      """
      return self.__selection
   
   def getBase(self):
      """
         Returns the current base directory, a string.
      """
      return self.__base
   
   def getFull(self):
      """
         Returns the current fully qualified selection.
      """
      return os.path.join(self.__base, self.__selection)

   def bind(self, seq, action):
      """
         Delegates bindings to the entry field.
      """
      return self.__entry.bind(seq, action)
   
   def __findIndex(self, name):
      """
         Finds the index of the simple name in current contents.
      """
      i = 0
      for file in self.__contents:
         if file.name == name:
            return i
         i += 1
   
   def __show(self):
      
      # add the selection to the entryfield
      self.__entry.delete(0, END)
      self.__entry.insert(0, os.path.join(self.__base, self.__selection))
      
      # find the directory of the selection
      self.__list.deleteAll()
      for file in self.__contents:
         self.__list.append(file.fmt(self.__format))
      
      # preselect the selection in the list
      if self.__selection:
         index = self.__findIndex(self.__selection)
         self.__list.list.selection_set(index)
         self.__list.list.see(index)

   def __setSelection(self, index):
      self.__list.list.selection_clear(0, 'end')
      self.__list.list.selection_set(index)
      self.__list.list.see(index)
      self.__selected(None)

   def __fixEntryView(self):
      "Shifts the entry view so that the right edge is always visible"
      left, right = self.__entry._getdoubles(self.__entry.tk.call((self.__entry._w, 'xview')))
      if right < 1.0:
         self.__entry.xview_moveto(1.0 - right)

   def __setEntry(self):
      "Sets the entry from the current `__selection` variable"
      self.__entry.delete(0, 'end')
      self.__entry.insert(0, os.path.join(self.__base, self.__selection))      
      self.__fixEntryView()

   def __getSelection(self):
      index = self.__list.firstSelectionIndex()
      if index is not None:
         return self.__contents[index]
      else:
         return None

   def __selected(self, event):
      "called when an item is selected in the listbox"
      if self.__modifyEntry:
         self.__selection = self.__getSelection().name
         self.__setEntry()

   def __reselect(self, partialFileName):
      "Reselect from the listbox based on the contents of the entry"
      self.__modifyEntry = False
      i = 0
      ok = True

      if partialFileName:
      
         # find the first match in the listbox
         for item in self.__contents:
            #print 'checking %s against %s' % (item, partialFileName)
            if item.name.startswith(partialFileName):
               #print 'matched %s' % item
               self.__setSelection(i)
               break
            i += 1
         else:
            # no match, if this is allowed, deselect everything.  Otherwise 
            # return an error
            if self.__canCreate:
               self.__list.list.selection_clear(0, 'end')
            else:
               ok = False
      
      else:
         # no selection, deselect all
         self.__list.list.selection_clear(0, 'end')
      
      self.__modifyEntry = True      
      return ok

   def __alert(self):
      "Beep at the user"
      self.bell()

   def __key(self, evt):
      print evt.keysym, evt.keycode, evt.char, evt.state #, dir(evt)
      
      # handle cursor up/down keys - we don't deal with changing
      # the contents of the entry, let the selection event take care of
      # that
      
      if evt.keysym == 'Up':
         index = self.__list.firstSelectionIndex()
         if index is None:
            # nothing currently selected, go to the end of the list
            index = len(self.__contents)
            if index:
               self.__setSelection(index - 1)
         elif index > 0:
            self.__setSelection(index - 1)
      elif evt.keysym == 'Down':
         index = self.__list.firstSelectionIndex()
         if index is None:
            # nothing currently selected, go to the beginning of the list
            if self.__contents:
               self.__setSelection(0)
         elif index < len(self.__contents) - 1:
            self.__setSelection(index + 1)
      elif evt.keysym == 'BackSpace':
         
         # if this is Ctrl-Backspace, we want to back up to the parent 
         # directory
         if evt.state & 4 == 4:
            if self.__selection:
               # clear the selection, leave the base
               self.set(os.path.join(self.__base, ''))
            elif len(self.__base) > 1:
               # remove the last directory segment of the base
               self.set(os.path.join(os.path.split(self.__base)[0], ''))
            self.__show()
            return 'break'
         
         # normal backspace functionality
         if self.__selection:
            # remove the last character from the selection
            self.__selection = self.__selection[:-1]
            self.__entry.delete(0, 'end')
            contents = os.path.join(self.__base, self.__selection)
            self.__entry.insert(0, contents)
            
            # Reselect from the contents of the entry
            print 'reselecting %s' % self.__selection
            self.__reselect(self.__selection)
         elif len(self.__base) > 1:
            self.set(self.__base)            
            self.__show()
            
         return 'break'

      elif evt.keysym == 'Tab':
         # get the first selection
         sel = self.__getSelection()
         if sel:
            # set the entry field from the selection
            self.__selection = sel.name
            self.__setEntry()
         else:
            self.__alert()
         return 'break'

      elif evt.keysym == 'slash':
         path = os.path.join(self.__base, self.__selection, '')
         if os.path.isdir(path):
            self.set(path)
         else:
            self.__alert()
         return 'break'

      elif evt.keysym == 'Next':
         self.__list.list.yview_scroll(1, PAGES)
      
      elif evt.keysym == 'Prior':
         self.__list.list.yview_scroll(-1, PAGES)

      elif evt.keysym[:1] in 'qw' and evt.state & 4 == 4:
         self.changeFormat({'q': 0, 'w': 1}[evt.keysym])

      elif evt.char and evt.char in self.__normalChars:
         # try adding the new character, make sure the results are acceptable
         # before adding the change to the entryfield
         sel = self.__selection + evt.char
         if self.__reselect(sel):
            self.__entry.insert('end', evt.char)
            self.__selection = sel
         else:
            self.__alert()
         return 'break'
      
      else:
         return None

   def focus(self):
      self.__entry.focus()

def quit(evt):
   top.destroy()

if __name__ == '__main__':
   top = Tk()
   top.bind('<Return>', quit)
   browser = FileBrowser(top, filter = GlobFilter(['*.py'], File.REG))
   browser.pack(expand = 1, fill = BOTH)
   browser.focus()
   browser.mainloop()
   print browser.isValid(), browser.get(), browser.getBase(), browser.getFull()
