#==============================================================================
#
#  $Id: phunkyvw.py 561 2000-02-13 22:35:04Z mike $
#
"""
   Self constructing views of objects.
"""
#
#  Copyright (C) 1999 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#  $Log: phunkyvw.py,v $
#  Revision 1.3  2000/02/13 22:35:04  mike
#  Doc fixes
#
#  Revision 1.2  1999/05/23 17:42:22  mike
#  Lots of changes - still not ready for prime time.
#
#  Revision 1.1  1999/04/08 01:01:35  mike
#
#  First cut at phunkyvw; on StdWin fixed docs and cleaned-up functionality.
#
#  Revision 1.1  1999/02/17 01:29:47  mike
#  Initial revision
#
#
#==============================================================================


"""
   todo:
   
      -  Add SmallTextViewer in all its forms
      -  Add getvalue() methods to ListViewer and DictViewer
      -  add dataChanged() method to InstViewer and dataChanged(start, end)
         to list viewer.

   Notes on ListViewer:
      Can be eitther:
          Completely transactional (in which case I need to begin a transaction
          when the window is opened and end it when it is closed)
          
             Could mean that I'd have to lock out all other transactions and
             modifications to the object.
             
             I don't think that having a transaction open necessarily 
             is exclusive.  The transaction is just a set of changes.  They
             can be applied to an existing object event while there are
             changes pending upon an object
             
             Can there be more than one change pending upon an object?
             If so, do we tolerate the race condition that results when 
             two transactions are open at the same time?
             
             I don't see why not.  I've already decided that I want to be 
             able to view changes to lists even as I make changes to lists.             
         
         End transactional (like an InstViewer), creating a transaction
         only during the getvalue() method.
            
            Means I'd have to save all of the actions performed on the list - 
            but that's what the transaction is for!
   
   Should define control components which can either themselves be GUI
   components or should contain GUI components (allowing a `label/entry`
   combination to be in two different grid cells!)
   
   Notes on attribute specialization:
      
      Want to be able to define attributes like this:
         
         ('attrName', 'Name on Label', ControlComponentClass, 
            
            # dictionary defining component information, including positioning
            # information.
            mkdict(attribute1 = 'val', ...)
          )
      
      So that a window allowing you to edit a persons name, address
      and a list of e-mail addresses could be indicated like so:
      
         [
         ('name', 'Full Name:', SmallTextViewer, mkdict(row = 0, column = 0)),
         ('fullAddress', 'Address:', FullTextViewer,
            mkdict(row = 1, column = 0, columnspan = 2),
          ),
         ('emailAddress', 'EMail Addresses:', StringListViewer,
            mkdict(row = 2, column = 0, columnspan = 2),
          ),
         ]

   Should be able to use the same description for the rendering of HTML forms
   using pycgi!

   Must provide event notification methods that are accessible from non-GUI 
   threads.
   
   Notes on desired components:
      
      List of strings manager:
         
         Contains a list of simple strings with add, edit, and delete buttons.
         Contains a built in entryfield for adding and editing items in the
         list.
         
         Should also have a variation that allows the user to use a choice
         control to add and edit.
   
"""

from types import *
from UserDict import UserDict
from Tkinter import *
from spug.win.Controls import ScrollingList, FullEntry
from spug.win.StdWin import MessageBox

class Utility:
   """
   
      The utility class simply supports setting the general case of setting
      instance attributes from a dictionary with defaults.
   
      Attribute types are:
      
      *bool*::
         Boolean.  "true" or "false", 1 or 0 in python code.
      *func*::
         Function.  Must be a callable object in python, not sure how
         this translates into a string.
      *class*::
         Class.  Must be a class object in python, the fully qualified
         class name (with its module) as a string.
      *str*::
         A string value.
      *int*::
         An integer value.
   """
   def _selectAttrs(self, dict, attrs):
      """
         Returns a new dictionary composed of the identified /attrs/. 
         /attrs/ should be a list of string keys and-or 2-tuples, 
         for each tuple, the first item is the key as it is known in /dict/,
         the second item is the key that the value will be entered under
         in the return dictionary.
      """
      newdict = {}
      for attr in attrs:
         if type(attr) is StringType and dict.has_key(attr):
            newdict[attr] = dict[attr]
         elif type(attr) is TupleType and dict.has_key(attr[0]):
            newdict[attr[1]] = dict[attr[0]]
      return newdict         
   
   def _setattrs(self, dict, attrs):
      """
         This insidious little method walks through /attrs/ (which should be
	 a list of tuples of attribute-name, attribute-type, and
	 default-value triplets) and sets the named attribute in /self/ to
	 the value associated with the key in the dictionary /dict/ (if it
	 exists) or to the default value if it does not.
         
         XXX function needs to go a step beyond this and read the information
         from the objects window attributes if it has not been specified
         in the dictionary.
      """
      self.__attrs = {}
      for key, type, default in attrs:
         val = dict.get(key, default)
         self.__attrs[key] = val
         setattr(self, key, val)

   def _attrs(self):
      """
         Returns the attribute dictionary for the object.
      """
      return self.__attrs

class ObjectEditor:
   
   """
      The base class for all components that can be used to edit objects at
      any level.  An abstract interface.
      
      Public Variables:
      
      /org/::
         The original object.  It is assumed that the original object can
         always be used as a source of reset information.
      
      /txn/::
         The new object.  It is assumed that it is always safe to set the 
         attributes on a transaction object and that the changes will not
         take effect until the object is committed.

      /renderer/::
         The physical renderer that will be used.
      
      Conceptually, an object editor manages the interaction between a 
      transaction (represented by /org/ and /txn/) and a concrete
      representation in `display/edit` rendering paradigm.
      
      All object editors should have the following rendering attributes:
      
      /row/::
         Row the control occupies within its grid.
      /column/::
         Column that the control occupies within its grid.
      /rowspan/::
         Number of rows that the control spans
      /columnspan/::
         Number of columns that the control spans.
      /sticky/::
         Indicates which side of it's cell the control sticks to.  Must be 
         a string composed of some combination of the letters 'n', 's', 'e',
         and 'w'.
   """

   def __init__(self, **attrs):
      """
      """
      self.org = None
      self.txn = None
      self.attrs = attrs

   def commit(self):
      """
         Commits the transaction that the editor represents, returning the
         new data object.
      """
      raise ImplementedBySubclassError()
   
   def validate(self):
      """
         Checks to make sure that all values changed within the editor are
         correct.
      """
      raise ImplementedBySubclassError()
   
   def reset(self):
      """
         Resets the value of the all of the fields in the object editor.
      """
      raise ImplementedBySubclassError()

   def getClassName(self):
      """
         Returns the class name used to map to the particular renderer
         for the object.
      """
      raise ImplementedBySubclassError()
   
   def renderOn(self, rendererFactory, parent, org, txn):
      """
         Renders the editor using the given instance of @RendererFactory.
         /parent/ is the parent component, which should also be a
         @Renderer.
      """
      self.org = org
      self.txn = txn
      self.renderer = \
         rendererFactory.create(self.getClassName(), parent, self, self.attrs)
   
   def generate(self, codeGenerator):
      """
         Generates a logical representation of the view using the given
         code generator.
      """
      raise ImplementedBySubclassError()

class Renderer:
   """
      A renderer is any kind of object that can be used to render a
      particular object view on a particular paradigm.
      
      This is the base class for all GUI objects that render the view.
   """
   
   def __init__(self, objectEditor):
      self._editor = objectEditor
   
   def activate(self):
      raise ImplementedBySubclassError()

class RendererFactory:
   
   """
      As the name implies, something that creates renderers.
   """
   
   def create(self, className):
      """
         Creates and returns a new renderer object.
      """
      raise ImplmentedBySubclassError()

class StringEditor(ObjectEditor):
   
   """
      A string editor should consist of a label and an entryfield allowing
      the editing of a single line of text.  The row and the column
      specified are normally for the label, the entryfield is one column to
      the right.  Therefore, this control normally occupies two cells.      
      
      Renderer for the string editor must contain the following methods:
      
      #get()#::
         Should return the string currently in the entryfield.
      
      #set(str)#::
         Should allow you to set the contents of the string editor.

      Attributes:
      
      *label*::
         The text of the label.
      
      *width*::
         The width of the entryfield in columns (only used for rendering,
         does not validate).

      *commitFunc*::
         The function that is called when the object is committed. *Not
         suppported at this time, commit method always returns a string.*
      
      The original value should be a string, the transaction object is
      ignored.
   """
   
   def __init__(self, **attrs):
      apply(ObjectEditor.__init__, (self,), attrs)   

   def getClassName(self):
      return 'StringEditor'
   
   def commit(self):
      return self.renderer.get()
   
   def validate(self):
      return 1
   
   def reset(self):
      self.renderer.set(self.org)

class Form(ObjectEditor):
   
   """
      A form is an editor for compound objects.

      Attributes:
      
   """
   
   def __init__(self, items, **attrs):
      apply(ObjectEditor.__init__, (self,), attrs)
      self.__child = {}
      for attr, klass, attrs in items:
         # create a new instance of the control and bind it 
         self.__child[attr] = apply(klass, (), attrs)

   def getClassName(self):
      return 'Form'
   
   def renderOn(self, factory, parent, obj, txn):
      ObjectEditor.renderOn(self, factory, parent, obj, txn)
      myRenderer = self.renderer
      for name, editor in self.__child.items():
         editor.renderOn(factory, myRenderer, 
                         name[0] != '.' and getattr(obj, name) \
                          or None,
                         name[0] != '.' and getattr(txn, name) \
                          or None
                         )
   
   def reset(self):
      for editor in self.__child.values():
         editor.reset()

   def validate(self):
      for editor in self.__child.values():
         if not editor.validate():
            return 0
      else:
         return 1
   
   def commit(self):
      for name, editor in self.__child.items():
         if name[0] != '.':
            setattr(self.txn, name, editor.commit())
      return self.txn
         
         
#=======================OLD STUFF========================================

class ButtonPanel(Frame, Utility):
   """
      Manages a row of buttons.
   """
   
   class EvtFunc:
      """
         Class that allows us to dispense with the event argument so that
         we can use the same action function for both the button and the
         key event.
      """
      
      def __init__(self, action):
         self.action = action
         
      def __call__(self, evt):
         self.action()
   
   def __init__(self, parent, **config):
      
      Frame.__init__(self, parent)
      self.buttons = []

   def addButton(self, name, action, key = None):
      """
         Adds the button specified by /name/ (a string button label)
         /action/ (the command to be invoked when the button is
         pressed) and /key/ (the key binding of the button).
         
         Returns the button object.
         
         Keys are bound to the toplevel widget.
      """
      btn = Button(self, text = name, command = action)
      btn.pack(side = LEFT)
      self.buttons.append(btn)
      if key:
         self.winfo_toplevel().bind(key, self.EvtFunc(action))
      return btn

def editObject(obj):
   panel = EditPanel()
   if hasattr(obj, 'phunkyEdit'):
      view = obj.phunkyEdit(panel, obj)
   elif type(obj) is ListType:
      view = ListViewer(panel, obj)
   elif type(obj) is DictType:
      view = DictViewer(panel, obj)
   elif type(obj) is InstanceType:
      view = InstViewer(panel, obj)
   else:
      view = MessageBox("Unable to view objects of this type")
      panel.destroy()
      return None

   panel.setView(view)
   return panel.getvalue()
   
class ListViewer(Frame, Utility):   
   """
      Attributes:
         /editable/::
            If true, the edit button is present. Set to true by default.
         /deletable/::
            If true, the *Delete*  button is present.  Set to true by
            default.
         /deleteItem/::
            Function that is called when the *Delete* button is pressed
            (by default, does a del /list/[itemIndex]).  The function should
            accept the data object and the index of the item to be deleted.
         /addItem/::
            Function that is called when the *Add* button is pressed.
            Should accept no parameters and return a new object.  This 
            function does not exist by default.
         /insertItem/::
            Function that is called when the *Edit* button is pressed.
            This function does not exist by default.
         /editItem/::
            Function that is called when the *Edit* button is pressed.  
            Should accept the object to be edited as its paramemeter, should
            return the new item, or *None* if the item was not changed.
            @editObject() is called by default.
         /itemAsStr/::
            Function that is called to obtain the string value of an item
            in the list.  #str()# by default.

         /beginEdit/::
            A function that is to be called before starting to set attributes
            in the original object.  Accepts the list as its parameter.
            By default, this is the @_beginEdit() method, which makes a copy
            of the list to be restored if we cancel.
         /endEdit/::
            The function that is to be called after setting attributes in the
            original object.  Accepts the instance as its parameter.
            *None* by default.
         /cancelEdit/::
            The function that is to be called when changes to the object
            have been cancelled.  Accepts the instance as its parameter.
            By default, this is the @_cancelEdit() method, which restores
            the list to its original value.
         
         The following items are unimplemented
         '''''''''''''''''''''''''''''''''''''
         
         /listFont/::
            Font used in the listbox.
         /buttonFont/::
            Font used in the buttons.
   """   
   
   def __init__(self, parent, data, **config):
      Frame.__init__(self, parent)

      self.data = data

      self._setattrs(config,
                     ( ('editable', 'bool', 1),
                       ('editItem', 'func', editObject),
                       ('deletable', 'bool', 1),
                       ('deleteItem', 'func', self._deleteItem),
                       ('addItem', 'func', None),
                       ('insertItem', 'func', None),
                       ('itemAsStr', 'func', str),
                       ('beginEdit', 'func', self._beginEdit),
                       ('endEdit', 'func', None),
                       ('cancelEdit', 'func', self._cancelEdit),
                      )
                     )
      
      self.list = ScrollingList(self)
      self.list.grid(row = 0, sticky = 'nsew')
      self.buttons = ButtonPanel(self)
      if self.addItem:
         self.buttons.addButton('Add', self.addPressed, '<Control-a>')
      if self.insertItem:
         self.buttons.addButton('Insert', self.insertPressed, '<Control-i>')
      if self.editable:
         self.buttons.addButton('Edit', self.editPressed, '<Control-e>')
      if self.deletable:
         self.buttons.addButton('Delete', self.deletePressed, '<Control-d>')      
      self.buttons.grid(row = 1, sticky = W)
      self.columnconfigure(0, weight = 1)
      self.rowconfigure(0, weight = 1)
      self.beginEdit(self.data)
      self.repopulate()

   def repopulate(self, start = 0, end = None):
      """
         Repopulates items /start/ to /end/ in the list.
      """
      
      # delete everything currently in those positions in the list
      if not end:
         self.list.delete(start, 'end')
      else:
         self.list.delete(start, end)
      
      # repopulate that portion of the list
      i = start
      for cur in self.data[start:]:
         self.list.insert(i, self.itemAsStr(cur))
         i = i + 1         
      
   def _deleteItem(self, itemIndex):
      "Internal method to delete an item."
      del self.data[itemIndex]

   def _beginEdit(self, list):
      "Internal method to begin editing."
      self.data0 = self.data[:]
   
   def _cancelEdit(self, list):
      "Internal method to restore the original list"
      self.data[:] = self.data0
   
   def getIndex(self):
      """
         Returns the index of the sected item.
      """
      index = self.list.firstSelectionIndex()
      if index is None:
         MessageBox("You must first select an item.").show()
      return index
   
   def editPressed(self):
      """
         Called when the edit button is pressed.
      """
      index = self.getIndex()
      if index is None:
         return
         
      newItem = apply(self.editItem, (self.data[index],))
      if newItem is None:
         return
         
      self.data[index] = newItem
      self.list.delete(index)
      self.list.insert(index, str(newItem))
   
   def deletePressed(self):
      """
         Called when the delete button is pressed.
      """
      index = self.getIndex()
      if index is None:
         return
      
      self.deleteItem(index)
      self.list.delete(index)

   def addPressed(self):
      """
         Called when the add button is pressed.
      """
      newItem = self.addItem()
      self.data.append(self.itemAsStr(newItem))

   def discardChanges(self):
      """
         Called to discard all changes that have been applied to the
         list.  Calls the cancelEdit attribute.
      """
      self.cancelEdit(self.data)

   def getvalue(self):
      """
      """         
      if self.endEdit:
         self.endEdit(self.data)
      return self.data
      
class DictAsList:
   """
      Representation of a dictionary as a list.  Provides __getitem__ 
      and __setitem__ so that the dictionary is now indexed by order.
   """
   
   def __init__(self, data):
   
      self.data = data
      self.list = data.items()
   
   def __getitem__(self, index):
      return self.list[index][1]
   
   def __setitem__(self, index, val):
      
      # modify the item in both the list and the dictionary
      key = self.list[index][1]
      self.list[index] = (key, val)
      setitem(self.data, key, val)
   
   def __delitem__(self, index):
   
      # delete the item in both the list and the dictionary.
      del self.data[self.list[index][0]]
      del self.list[index]
   
   def __getslice__(self, start, end):
      return self.list[start:end]
   
   def __setslice__(self, start, end, slice):
      for key, val in self.list[start:end]:
         del self.data[key]
      self.list[start:end] = slice
      for key, val in slice:
         self.data[key] = val


class DictViewer(ListViewer):
   """
      A dictionary viewer is currently a list viewer that deals with the
      keys of the dictionary.  It provides different defaults for the 
      following attributes:
      
      /editItem/::
         Default value constructs an editor for the associated value.
         XXX think we might want to allow them to edit the key.
   """
   def __init__(self, parent, data, **config):
      apply(ListViewer.__init__, (self, parent, DictAsList(data)), config)

class InstViewer(Frame, Utility):
   """
      Used to view an instance.
      
      Attributes:
      
      /beginEdit/::
         A function that is to be called before starting to set attributes
         in the original object.  Accepts the instance as its parameter.
         *None* by default.
      /endEdit/::
         The function that is to be called after setting attributes in the
         original object.  Accepts the instance as its parameter.
         *None* by default.
      /cancelEdit/::
         The function that is to be called when changes to the object
         have been cancelled.  Accepts the instance as its parameter.
         *None* by default.

   """

   class EditCmd:
      def __init__(self, editFunc, object, field):
         self.editFunc = editFunc
         self.object = object
         self.field = field
      
      def __call__(self, *parms):
         self.field.data = self.editFunc(self.object)

   class Field:
      def __init__(self, attr, data, viewer):
         self.attrName = attr
         self.data = data
         self.viewer = viewer
      
      def validate(self):
         return self.viewer.validate()
      
      def getvalue(self):
         return self.viewer.getvalue()

   class ButtonField(Field):

      def validate(self):
         return 1
      
      def getvalue(self):
         return self.data
   
   def __init__(self, parent, data, **config):
      Frame.__init__(self, parent)      
      self._setattrs(config,
                     [ ('beginEdit', 'func', None),
                       ('endEdit', 'func', None),
                       ('cancelEdit', 'func', None),
                       ]
                     )
      
      self.data = data
      
      # create a nice label on top
      self.className = Label(self, 
                             text = 'Instance of ' + data.__class__.__name__
                             )
      self.className.grid(row = 0, sticky = W)

      # the list where we will store the field objects that we crete
      self.fields = []
      
      # build a row for each instance
      row = 1
      
      for attrName, val in data.__dict__.items():
         if type(val) is StringType:
            label = Label(self, text = attrName, name = attrName + '_lbl')
            label.grid(row = row, column = 0, sticky = W)
            text = StringViewer(self, val)
            text.grid(row = row, column = 1, sticky = W)
            field = self.Field(attrName, val, text)
         elif type(val) in (IntType, LongType, FloatType):
            label = Label(self, text = attrName, name = attrName + '_lbl')
            label.grid(row = row, column = 0, sticky = W)
            text = IntViewer(self, val)
            text.grid(row = row, column = 1, sticky = W)
            field = self.Field(attrName, val, text)
         else:
            label = Label(self, text = attrName, name = attrName + '_lbl')
            label.grid(row = row, column = 0, sticky = W)
            field = self.ButtonField(attrName, val, None)
            btn = Button(self, text = type(val).__name__, 
                         command = self.EditCmd(editObject, val, field)
                         )
            btn.grid(row = row, column = 1, sticky = W)
         self.fields.append(field)
         row = row + 1

   def discardChanges(self):
      pass

   def getvalue(self):
      try:
         if self.beginEdit:
            self.beginEdit(self.data)
         for field in self.fields:
            setattr(self.data, field.attrName, field.getvalue())
         if self.endEdit:
            self.data = self.endEdit(self.data)
         return self.data
      finally:
         if self.cancelEdit:
            self.cancelEdit(self.data)

class SmallTextViewer(Entry):
   """
      A single field viewer/editor for editing small text.
   """
   def __init__(self, parent, text, **config):
      apply(Entry.__init__, (self, parent), config)
      self.insert(0, text)

class IntViewer(SmallTextViewer):
   """
      Allows us to view integers.
   """
   def __init__(self, parent, int, **config):      
      apply(SmallTextViewer.__init__, (self, parent, str(int)), config)
   
   def validate(self):
      try:
         int(self.get())
      except:
         return 0
      return 1
   
   def getvalue(self):
      return int(self.get())      

class StringViewer(SmallTextViewer):
   """
      Allows us to view integers.
   """
   def __init__(self, parent, str, **config):      
      apply(SmallTextViewer.__init__, (self, parent, str), config)

   def validate(self):
      return 1
   
   def getvalue(self):
      return self.get()

class EditPanel(Frame):
   """
      Has a "save" and "cancel" button at the bottom: allows you to edit
      views.
   """
   
   def __init__(self, parent = None, **config):
      if not parent:
         parent = Toplevel()
      self.top = parent
      apply(Frame.__init__, (self, parent,), config)
      self.view = None
      self.buttons = ButtonPanel(self)
      self.buttons.addButton('Save', self.savePressed, '<Return>')
      self.buttons.addButton('Cancel', self.cancelPressed, '<Escape>')
      self.buttons.grid(row = 1, sticky = W)
      self.columnconfigure(0, weight = 1)
      self.rowconfigure(0, weight = 1)
      self.pack(expand = 1, fill = 'both')
      self.data = None
   
   def savePressed(self):
      self.data = self.view.getvalue()
      self.top.destroy()

   def cancelPressed(self):
      self.view.discardChanges()
      self.top.destroy()
   
   def setView(self, view):
      self.view = view
      self.view.grid(row = 0, column = 0, sticky = 'nsew')
   
   def getvalue(self):
      self.wait_window(self)
      return self.data
     
if __name__ == '__main__':
   tk = Tk()
   edit = Button(tk, text = 'Edit Object', 
                 command = lambda: editObject([1, 2, 3])
                 )
   edit.pack()
   tk.mainloop()
      
