/*===========================================================================*\

   $Id$

   Copyright (C) 2007 Michael A. Muller

   Permission is granted to use, modify and redistribute this code,
   providing that the following conditions are met:
 
   1) This copyright/licensing notice must remain intact.
   2) If the code is modified and redistributed, the modifications must 
   include documentation indicating that the code has been modified.
   3) The author(s) of this code must be indemnified and held harmless
   against any damage resulting from the use of this code.

   This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
   warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

\*===========================================================================*/

#include "WurldSDLAudio.h"

#include <SDL.h>
#include <SDL_mixer.h>
#include <list>
#include <spug/types.h>
#include <spug/StringFmt.h>
#include <spug/Exception.h>
#include <spug/Locker.h>
#include <spug/InvalidStateException.h>

using namespace spug;

namespace {
   SPUG_RCPTR(MySample);

   class MySample : public Sample {
      public:
         Mix_Chunk *chunk;

         MySample(const char *filename) {
            chunk = Mix_LoadWAV(filename);
            if (!chunk)
               throw Exception(SPUG_FSTR("Failed to load " << filename));
         }

   };


};

class WurldSDLAudioPlayer : public AudioPlayer {
   private:
      WurldSDLAudio *driver;
      MySamplePtr sample;
      MoveablePtr moveable;
      Vector position;
      int channel;

      // maximum audible distance = 100 meters
      static const float MAX_AUDIO_DISTANCE = 1000;
      
   public:
      WurldSDLAudioPlayer(WurldSDLAudio *driver, const SamplePtr &sample) :
         driver(driver),
         channel(-1) {
         this->sample = MySamplePtr::dcast(sample);
      }

      virtual void start() {
         if (channel != -1)
            throw InvalidStateException("Player is already playing");

         channel = Mix_PlayChannel(-1, sample->chunk, 0);
         driver->addPlayer(channel, this);
      }

      virtual void stop() {
         if (channel == -1)
            throw InvalidStateException("Player is not playing");

         Mix_HaltChannel(channel);
         channel = -1;
      }

      virtual void setMoveable(const MoveablePtr &moveable) {
         this->moveable = moveable;
      }

      void changePosition(const Camera &listener) {
         if (moveable) {

            // get the relative position of the audio source
            Vector relPos = moveable->getPosition() - listener.getOrigin();

            // rotate it into the perspective of the listener
            RotMatrix rotation;
            rotation.makeInverse(listener.getRotation());
            Vector rotRelPos;
            rotation.rotate(relPos, rotRelPos);
            //std::cout << "rel pos = " << relPos << " rot rel pos = " 
            //   << rotRelPos << std::endl;


            // calculate the direction of the sound as an angle around the
            // y axis
            Float angle;
            if (rotRelPos.z > 0)
               // for positive z (behind us) we want +/-90 + the angle from
               // the X axis
               angle = copysign(90, rotRelPos.x) + 
                       degrees(atan(rotRelPos.z / rotRelPos.y));
            else
               // for negative z, we want angle from the z axis
               angle = degrees(atan(rotRelPos.x / -rotRelPos.z));

            // calculate the distance as a fraction from 0 to 1.  1 is the
            // maximum distance, 0 is the closest.
            float dist = relPos.size() / MAX_AUDIO_DISTANCE;
            if (dist > 1)
               dist = 1;

            //std::cout << "channel = " << channel << " angle = " << angle
            //   << " dist = " << dist << std::endl;
            Mix_SetPosition(channel, (Sint16)angle, (Uint8)(dist * 256));
         } else {
            // XXX if moveable is null and was null, do nothing
            Mix_SetPosition(0, 0, 0);
         }
      }

      virtual void handleStopped() {
         std::cout << "finished playing" << std::endl;
      }

};

WurldSDLAudio *WurldSDLAudio::instance = 0;

extern "C" {
   static void wurldChannelCallback(int channel) {
      WurldSDLAudio::getInstance()->handlePlayerStopped(channel);
   }
}

WurldSDLAudio::WurldSDLAudio() {
   if (SDL_Init(SDL_INIT_AUDIO) < 0)
      throw Exception(SPUG_FSTR("Unable to initialize SDL Audio: "
                                 << SDL_GetError()
                                )
                      );
   if (Mix_OpenAudio(44100, AUDIO_S16, 2, 4096) < 0)
      throw Exception(SPUG_FSTR("Unable to open audio device: " 
                                 << SDL_GetError()
                                )
                      );

   // set the callback
   Mix_ChannelFinished(wurldChannelCallback);

   instance = this;
}

WurldSDLAudio::~WurldSDLAudio() {
   instance = 0;
}

SamplePtr WurldSDLAudio::createSampleFromFile(const char *filename) {
   return new MySample(filename);
}

AudioPlayerPtr WurldSDLAudio::createPlayer(const SamplePtr &sample) {
   return new WurldSDLAudioPlayer(this, sample);
}

void WurldSDLAudio::addPlayer(int channel, const AudioPlayerPtr &player) {
   Locker lock(mapLock);
   players[channel] = player;
}

void WurldSDLAudio::process(const Camera &listener, long diff) {

   // create a list of all players (we can't do changePosition() with the
   // mutex locked or we risk deadlock)
   std::list<WurldSDLAudioPlayerPtr> holder;
   {
      Locker lock(mapLock);

      PlayerMap::iterator iter;
      for (iter = players.begin(); iter != players.end(); ++iter)
         holder.push_back(iter->second);
   }

   // set the distance and direction of all players
   for (std::list<WurldSDLAudioPlayerPtr>::iterator iter = holder.begin();
        iter != holder.end();
        ++iter
        )
      (*iter)->changePosition(listener);
}

void WurldSDLAudio::handlePlayerStopped(int channel) {
   Locker lk(mapLock);
   PlayerMap::iterator playerIter = players.find(channel);
   if (playerIter != players.end()) {
      playerIter->second->handleStopped();
      players.erase(playerIter);
   }
}
